use std::{convert::Infallible, future::Future, net::{IpAddr, SocketAddr}, pin::Pin};

use hyper::{body::Incoming, header::HeaderValue, server::conn::http1, service::Service as ServiceTrait, Request, Response, StatusCode};
use tokio::net::TcpListener;
use hyper_util::rt::{TokioIo, TokioTimer};

struct Service {
    pub remote_ip: Option<SocketAddr>
}

impl ServiceTrait<Request<Incoming>> for Service {
    type Response = Response<String>;
    type Error = Infallible;
    type Future = Pin<Box<dyn Future<Output = Result<Self::Response, Self::Error>> + Send>>;

    fn call(&self, req: Request<Incoming>) -> Self::Future {
        let mut remote_ip = self.remote_ip.map(|x| x.ip());

        if let Some(ip) = req.headers().get("X-Forwarded-For") {
            if let Ok(str_ip) = ip.to_str() {
                if let Ok(ip) = str_ip.parse::<SocketAddr>() {
                    remote_ip = Some(ip.ip());
                }
                if let Ok(ip) = str_ip.parse::<IpAddr>() {
                    remote_ip = Some(ip);
                }
            }
        }

        if let Some(ip) = remote_ip {
            let mut res = Response::new(ip.to_string());
            res.headers_mut().append("Content-Type", HeaderValue::from_static("text/plain"));
            Box::pin(async { Ok(res) })
        } else {
            let mut res = Response::new("no ip".to_string());
            res.headers_mut().append("Content-Type", HeaderValue::from_static("text/plain"));
            *res.status_mut() = StatusCode::BAD_REQUEST;
            
            Box::pin(async { Ok(res) })
        }
    }
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    let addr: SocketAddr = ([0, 0, 0, 0], 80).into();
    let listener = TcpListener::bind(addr).await?;
    println!("Listening on {addr:?}");

    loop {
        let (tcp, _) = listener.accept().await?;
        let remote_ip = tcp.peer_addr().map(|x| Some(x)).unwrap_or(None);
        let io = TokioIo::new(tcp);
        tokio::task::spawn(async move {
            let service = Service {
                remote_ip
            };
            if let Err(err) = http1::Builder::new()
                .timer(TokioTimer::new())
                .serve_connection(io, service).await {
                    eprintln!("{err:#?}");
                }
        });
    }
}
